import os

from vit import env


def get_xdg_config_dir(user_config_dir, resource):
    xdg_config_home = env.user.get("XDG_CONFIG_HOME") or os.path.join(
        os.path.expanduser("~"), ".config"
    )

    xdg_config_dirs = [xdg_config_home] + (
        env.user.get("XDG_CONFIG_DIRS") or "/etc/xdg"
    ).split(":")

    for config_dir in xdg_config_dirs:
        path = os.path.join(config_dir, resource)
        if os.path.exists(path):
            return path
    return None



from importlib import import_module
from datetime import datetime, timedelta
try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

from vit import util
from vit import uda
from vit.util import unicode_len

INDICATORS = [
    'active',
    'dependency',
    'recurrence',
    'tag',
]
UDA_DEFAULT_INDICATOR = 'U'

DEFAULT_DESCRIPTION_TRUNCATE_LEN=20

class FormatterBase:
    def __init__(self, loader, config, task_config, markers, task_colorizer):
        self.loader = loader
        self.config = config
        self.task_config = task_config
        self.markers = markers
        self.task_colorizer = task_colorizer
        self.date_default = self.task_config.translate_date_markers(self.task_config.subtree('dateformat')["default"])
        self.report = self.task_config.translate_date_markers(self.task_config.subtree('dateformat.report')) or self.date_default
        self.annotation = self.task_config.translate_date_markers(self.task_config.subtree('dateformat.annotation')) or self.date_default
        self.description_truncate_len = DEFAULT_DESCRIPTION_TRUNCATE_LEN
        self.epoch_datetime = datetime(1970, 1, 1, tzinfo=ZoneInfo('UTC'))
        self.due_days = int(self.task_config.subtree('due'))
        self.none_label = config.get('color', 'none_label')
        self.build_indicators()

    def build_indicators(self):
        for indicator in INDICATORS:
            label = self.task_config.subtree('%s.indicator' % indicator)
            setattr(self, 'indicator_%s' % indicator, label)
        self.indicator_uda = {}
        for uda_name in uda.get_configured(self.task_config).keys():
            label = self.task_config.subtree('uda.%s.indicator' % uda_name) or UDA_DEFAULT_INDICATOR
            self.indicator_uda[uda_name] = label

    def get_module_class_from_parts(self, parts):
        formatter_module_name = '_'.join(parts)
        formatter_class_name = util.file_to_class_name(formatter_module_name)
        return formatter_module_name, formatter_class_name

    def get_user_formatter_class(self, parts):
        formatter_module_name, formatter_class_name = self.get_module_class_from_parts(parts)
        return self.loader.load_user_class('formatter', formatter_module_name, formatter_class_name)

    def get_formatter_class(self, parts):
        formatter_module_name, formatter_class_name = self.get_module_class_from_parts(parts)
        try:
            formatter_module = import_module('vit.formatter.%s' % formatter_module_name)
            formatter_class = getattr(formatter_module, formatter_class_name)
            return formatter_class
        except ImportError:
            return None

    def get(self, column_formatter):
        parts = column_formatter.split('.')
        name = parts[0]

        formatter_class = self.get_user_formatter_class(parts)
        if formatter_class:
            return name, formatter_class

        formatter_class = self.get_formatter_class(parts)
        if formatter_class:
            return name, formatter_class

        uda_metadata = uda.get(name, self.task_config)
        if uda_metadata:
            is_indicator = parts[-1] == 'indicator'
            if is_indicator:
                formatter_class = self.get_formatter_class(['uda', 'indicator'])
                return name, formatter_class
            else:
                uda_type = uda_metadata['type'] if 'type' in uda_metadata else 'string'
                formatter_class = self.get_formatter_class(['uda', uda_type])
                if formatter_class:
                    return name, formatter_class
        return name, Formatter

    def format_subproject_indented(self, project_parts):
        if len(project_parts) == 1:
            subproject = project_parts[0]
            return (unicode_len(subproject), '', '', subproject)
        else:
            subproject = project_parts.pop()
            space_padding = (len(project_parts) * 2) - 1
            indicator = u'\u21aa '
            width = space_padding + unicode_len(indicator) + unicode_len(subproject)
            return (width, ' ' * space_padding , indicator, subproject)

    def recalculate_due_datetimes(self):
        self.now = datetime.now().astimezone()
        # NOTE: For some reason using self.zone for the tzinfo below results
        # in the tzinfo object having a zone of 'LMT', which is wrong. Using
        # the tzinfo associated with self.now returns the correct value, no
        # idea why this glitch happens.
        self.end_of_day = datetime(self.now.year, self.now.month, self.now.day, 23, 59, 59, tzinfo=self.now.tzinfo)
        self.due_soon = self.end_of_day + timedelta(days=self.due_days)

    def get_due_state(self, due, task):
        if util.task_pending(task):
            if due < self.now:
                return 'overdue'
            elif due <= self.end_of_day:
                return 'due.today'
            elif due < self.due_soon:
                return 'due'
        return None

    def get_active_state(self, start, task):
        end = task['end']
        return util.task_pending(task) and start and start <= self.now and (not end or end > self.now)

    def get_scheduled_state(self, scheduled, task):
        return scheduled and not util.task_completed(task)

    def get_until_state(self, until, task):
        return until and not util.task_completed(task)





#!/usr/bin/env python

from importlib import import_module

import os
import signal
import subprocess
# TODO: Use regex module for better PCRE support?
#       https://bitbucket.org/mrabarnett/mrab-regex
import re
import time
import copy
from inspect import isfunction
from functools import reduce

import urwid

from vit import version
from vit.exception import VitException
from vit.formatter_base import FormatterBase
from vit import event
from vit.loader import Loader
from vit.config_parser import ConfigParser, TaskParser
from vit.util import clear_screen, string_to_args, is_mouse_event, task_id_or_uuid_short
from vit.process import Command
from vit.task import TaskListModel
from vit.autocomplete import AutoComplete
from vit.keybinding_parser import KeybindingParser
from vit.help import Help
from vit.key_cache import KeyCache
from vit.actions import Actions
from vit.markers import Markers
from vit.color import TaskColorConfig, TaskColorizer
from vit.task_list import TaskTable
from vit.multi_widget import MultiWidget
from vit.command_bar import CommandBar
from vit.registry import ActionRegistry, RequestReply
from vit.action_manager import ActionManagerRegistry
from vit.denotation import DenotationPopupLauncher
from vit.pid_manager import PidManager

# NOTE: This entire class is a workaround for the fact that urwid catches the
# 'ctrl l' keypress in its unhandled_input code, and prevents that from being
# used for the refresh report functionality. Sadly, that is the default
# keybinding, therefore the only way to make it work is to catch the refresh
# action here in the top frame.
class MainFrame(urwid.Frame):
    def __init__(self, body, header=None, footer=None, focus_part='body', key_cache=None, action_manager=None, refresh=None):
        self.key_cache = key_cache
        self.action_manager = action_manager
        self.refresh = refresh
        self.register_managed_actions()
        super().__init__(body, header, footer, focus_part)

    def register_managed_actions(self):
        self.action_manager_registrar = self.action_manager.get_registrar()
        self.action_manager_registrar.register('REFRESH', self.refresh)

    def is_default_refresh_key(self, keys):
        return keys == 'ctrl l'

    def keypress(self, size, key):
        keys = self.key_cache.get(key)
        if self.action_manager_registrar.handled_action(keys) and self.is_default_refresh_key(keys):
            # NOTE: Calling refresh directly here to avoid the
            # action-manager:action-executed event, which clobbers the load
            # time currently.
            self.refresh()
            return None
        else:
            return super().keypress(size, key)

class Application:
    def __init__(self, option, filters):
        self.extra_filters = filters
        self.loader = Loader()
        self.load_early_config()
        self.set_report()
        self.setup_main_loop()
        self.setup_signal_listeners()
        self.refresh(False)
        self.setup_pid()
        self.loop.run()

    def setup_signal_listeners(self):
        # Since not all platforms may have all signals, ensure they are
        # supported before adding a handler.
        if hasattr(signal, 'SIGUSR1'):
            pipe = self.loop.watch_pipe(self.async_refresh)
            def sigusr1_handler(signum, frame):
                os.write(pipe, b'x')
            signal.signal(signal.SIGUSR1, sigusr1_handler)
        if hasattr(signal, 'SIGTERM'):
            def sigterm_handler(signum, frame):
                self.signal_quit("SIGTERM")
            signal.signal(signal.SIGTERM, sigterm_handler)
        if hasattr(signal, 'SIGINT'):
            def sigint_handler(signum, frame):
                self.signal_quit("SIGINT")
            signal.signal(signal.SIGINT, sigint_handler)
        if hasattr(signal, 'SIGQUIT'):
            def sigquit_handler(signum, frame):
                self.signal_quit("SIGQUIT")
            signal.signal(signal.SIGQUIT, sigquit_handler)

    def load_early_config(self):
        self.config = ConfigParser(self.loader)
        self.task_config = TaskParser(self.config)
        self.reports = self.task_config.get_reports()

    def set_report(self):
        if len(self.extra_filters) == 0:
            self.report = self.config.get('report', 'default_report')
        elif self.extra_filters[0] in self.reports:
            self.report = self.extra_filters.pop(0)
        else:
            self.report = self.config.get('report', 'default_filter_only_report')

    def setup_main_loop(self):
        self.loop = urwid.MainLoop(urwid.Text(''), unhandled_input=self.key_pressed, pop_ups=True, handle_mouse=self.config.mouse_enabled)
        try:
            self.loop.screen.set_terminal_properties(colors=256)
        except:
            pass

    def setup_pid(self):
        self.pid_manager = PidManager(self.config)
        self.pid_manager.setup()

    def teardown_pid(self):
        self.pid_manager.teardown()

    def set_active_context(self):
        self.context = self.task_config.get_active_context()

    def load_contexts(self):
        self.contexts = self.task_config.get_contexts()

    def bootstrap(self, load_early_config=True):
        self.loader = Loader()
        if load_early_config:
            self.load_early_config()
        self.load_contexts()
        self.set_active_context()
        self.event = event.Emitter()
        self.setup_config()
        self.search_term_active = ''
        self.search_direction_reverse = False
        self.action_registry = ActionRegistry()
        self.actions = Actions(self.action_registry)
        self.actions.register()
        self.keybinding_parser = KeybindingParser(self.loader, self.config, self.action_registry)
        self.command = Command(self.config)
        self.get_available_task_columns()
        self.setup_keybindings()
        self.action_manager = ActionManagerRegistry(self.action_registry, self.key_cache.keybindings, event=self.event)
        self.register_managed_actions()
        self.markers = Markers(self.config, self.task_config)
        self.theme = self.init_theme()
        self.theme_alt_backgrounds = self.get_theme_alt_backgrounds()
        self.task_color_config = TaskColorConfig(self.config, self.task_config, self.theme, self.theme_alt_backgrounds)
        self.init_task_colors()
        self.task_colorizer = TaskColorizer(self.task_color_config)
        self.formatter = FormatterBase(self.loader, self.config, self.task_config, self.markers, self.task_colorizer)
        self.request_reply = RequestReply()
        self.set_request_callbacks()
        # TODO: TaskTable is dependent on a bunch of setup above, this order
        # feels brittle.
        self.build_task_table()
        self.help = Help(self.keybinding_parser, self.actions.get(), event=self.event, request_reply=self.request_reply, action_manager=self.action_manager)
        self.event.listen('command-bar:keypress', self.command_bar_keypress)
        self.event.listen('task:denotate', self.denotate_task)
        self.event.listen('action-manager:action-executed', self.action_manager_action_executed)
        self.event.listen('help:exit', self.deactivate_help)

    def setup_config(self):
        self.confirm = self.config.confirmation_enabled
        self.wait = self.config.wait_enabled

    def register_managed_actions(self):
        # Global.
        self.action_manager_registrar = self.action_manager.get_registrar()
        self.action_manager_registrar.register('QUIT', self.quit)
        self.action_manager_registrar.register('QUIT_WITH_CONFIRM', self.activate_command_bar_quit_with_confirm)
        # NOTE: This is a no-op for the default refresh keybinding, which is
        # handled by the MainFrame() class. It's included here in case the user
        # assigns a non-default key to the REFRESH action.
        self.action_manager_registrar.register('REFRESH', self.refresh)
        self.action_manager_registrar.register('TASK_ADD', self.activate_command_bar_add)
        self.action_manager_registrar.register('REPORT_FILTER', self.activate_command_bar_filter)
        self.action_manager_registrar.register('TASK_UNDO', self.task_undo)
        self.action_manager_registrar.register('TASK_SYNC', self.task_sync)
        self.action_manager_registrar.register('COMMAND_BAR_EX', self.activate_command_bar_ex)
        self.action_manager_registrar.register('COMMAND_BAR_EX_TASK_READ_WAIT', self.activate_command_bar_ex_read_wait_task)
        self.action_manager_registrar.register('COMMAND_BAR_SEARCH_FORWARD', self.activate_command_bar_search_forward)
        self.action_manager_registrar.register('COMMAND_BAR_SEARCH_REVERSE', self.activate_command_bar_search_reverse)
        self.action_manager_registrar.register('COMMAND_BAR_SEARCH_NEXT', self.activate_command_bar_search_next)
        self.action_manager_registrar.register('COMMAND_BAR_SEARCH_PREVIOUS', self.activate_command_bar_search_previous)
        self.action_manager_registrar.register('COMMAND_BAR_TASK_CONTEXT', self.activate_command_bar_task_context)
        self.action_manager_registrar.register('GLOBAL_ESCAPE', self.global_escape)
        self.action_manager_registrar.register(self.action_registry.noop_action_name, self.action_registry.noop)
        # Task.
        self.action_manager_registrar.register('TASK_ANNOTATE', self.task_action_annotate)
        self.action_manager_registrar.register('TASK_DELETE', self.task_action_delete)
        self.action_manager_registrar.register('TASK_DENOTATE', self.task_action_denotate)
        self.action_manager_registrar.register('TASK_MODIFY', self.task_action_modify)
        self.action_manager_registrar.register('TASK_START_STOP', self.task_action_start_stop)
        self.action_manager_registrar.register('TASK_DONE', self.task_action_done)
        self.action_manager_registrar.register('TASK_PRIORITY', self.task_action_priority)
        self.action_manager_registrar.register('TASK_PROJECT', self.task_action_project)
        self.action_manager_registrar.register('TASK_TAGS', self.task_action_tags)
        self.action_manager_registrar.register('TASK_WAIT', self.task_action_wait)
        self.action_manager_registrar.register('TASK_EDIT', self.task_action_edit)
        self.action_manager_registrar.register('TASK_SHOW', self.task_action_show)

    def default_keybinding_replacements(self):
        import json
        from datetime import datetime
        task_replacement_match = re.compile(r"^TASK_(\w+)$")
        def _task_attribute_match(variable):
            matches = re.match(task_replacement_match, variable)
            if matches:
                attribute = matches.group(1).lower()
                if attribute in self.available_columns:
                    return [attribute]
        def _task_attribute_replace(task, attribute):
            if task and task[attribute]:
                if type(task[attribute]) in ['set', 'tuple', 'dict', 'list']:
                    try:
                        json.dumps(task[attribute])
                    except Exception as e:
                        raise RuntimeError('Error parsing task attribute %s as JSON: %s' % (task[attribute], e))
                elif isinstance(task[attribute], datetime):
                    return task[attribute].strftime(self.formatter.report)
                else:
                    return str(task[attribute])
            return ''
        replacements = [
            {
                'match_callback': _task_attribute_match,
                'replacement_callback': _task_attribute_replace,
            },
        ]
        return replacements

    def add_user_keybinding_replacements(self, replacements):
        klass = self.loader.load_user_class('keybinding', 'keybinding', 'Keybinding')
        if klass:
            keybinding_custom = klass()
            replacements.extend(keybinding_custom.replacements())
        return replacements

    def wrap_replacements_callbacks(self, replacements):
        def build_wrapper(callback):
            def wrapper(*args):
                _, task = self.get_focused_task()
                return callback(task, *args)
            return wrapper
        for i, replacement in enumerate(replacements):
            replacements[i]['replacement_callback'] = build_wrapper(replacement['replacement_callback'])
        return replacements

    def setup_keybindings(self):
        self.keybinding_parser.load_default_keybindings()
        bindings = self.config.items('keybinding')
        replacements = self.wrap_replacements_callbacks(self.add_user_keybinding_replacements(self.default_keybinding_replacements()))
        keybindings = self.keybinding_parser.add_keybindings(bindings=bindings, replacements=replacements)
        self.key_cache = KeyCache(keybindings)
        self.key_cache.build_multi_key_cache()

    def set_request_callbacks(self):
        self.request_reply.set_handler('application:keybindings', 'Get keybindings', lambda: self.keybinding_parser.keybindings)
        self.request_reply.set_handler('application:key_cache', 'Get key cache', lambda: self.key_cache)
        self.request_reply.set_handler('application:blocking_task_uuids', 'Get blocking task uuids', lambda: self.blocking_task_uuids)

    def init_theme(self):
        theme = self.config.get('vit', 'theme')
        user_theme = self.loader.load_user_class('theme', theme, 'theme')
        if user_theme:
            return user_theme
        try:
            return import_module('vit.theme.%s' % theme).theme
        except ImportError:
            raise ImportError("theme '%s' not found" % theme)

    def get_theme_setting(self, setting):
        for s in self.theme:
            if s[0] == setting:
                return s

    def get_theme_alt_backgrounds(self):
        stiped_table_row = self.get_theme_setting('striped-table-row')
        return {
            '.striped-table-row': (stiped_table_row[2], stiped_table_row[5]),
        }

    def init_task_colors(self):
        self.theme += self.task_color_config.display_attrs

    def clear_key_cache(self):
        self.key_cache.set()
        self.update_status_key_cache()

    def action_manager_action_executed(self, data):
        self.clear_key_cache()

    def check_macro(self, keys):
        keybinding = self.keybinding_parser.keybindings[keys] if keys in self.keybinding_parser.keybindings else False
        return keybinding if keybinding and 'keys' in keybinding else False

    def execute_macro(self, keys):
        keybinding = self.check_macro(keys)
        if keybinding:
            keypresses = self.prepare_keybinding_keypresses(keybinding['keys'])
            self.loop.process_input(keypresses)

    def prepare_keybinding_keypresses(self, keypresses):
        def reducer(accum, key):
            if type(key) is tuple:
                accum += list(key[0](*key[1]))
            else:
                accum.append(key)
            return accum
        return reduce(reducer, keypresses, [])

    def denotate_task(self, data):
        task = self.model.task_denotate(data['uuid'], data['annotation'])
        if task:
            self.table.flash_focus()
            self.update_report()
            self.activate_message_bar('Task %s denotated' % self.model.task_id(task['uuid']))
            self.task_list.focus_by_task_uuid(data['uuid'], self.previous_focus_position)

    def command_bar_keypress(self, data):
        metadata = data['metadata']
        op = metadata['op']
        if 'choices' in data['metadata']:
            choice = data['choice']
            if op == 'quit' and choice:
                self.quit()
            elif op == 'done' and choice is not None:
                self.task_done(metadata['uuid'])
            elif op == 'delete' and choice is not None:
                self.task_delete(metadata['uuid'])
            elif op == 'start-stop' and choice is not None:
                self.task_start_stop(metadata['uuid'])
            elif op == 'priority' and choice is not None:
                task = self.model.task_priority(metadata['uuid'], choice)
                if task:
                    self.table.flash_focus()
                    self.update_report()
                    self.activate_message_bar('Task %s priority set to: %s' % (self.model.task_id(task['uuid']), task['priority'] or 'None'))
        elif data['key'] in ('enter',):
            args = string_to_args(data['text'])
            if op == 'ex':
                metadata = self.ex(data['text'], data['metadata'])
            elif op == 'filter':
                self.extra_filters = args
                self.update_report()
            elif op == 'project':
                # TODO: Validation if more than one arg passed.
                project = args[0] if len(args) > 0 else ''
                task = self.model.task_project(metadata['uuid'], project)
                if task:
                    self.table.flash_focus()
                    self.update_report()
                    self.activate_message_bar('Task %s project updated' % self.model.task_id(task['uuid']))
            elif op == 'wait':
                # TODO: Validation if more than one arg passed.
                wait = args[0] if len(args) > 0 else ''
                # NOTE: Modify is used here to support the special date
                # handling Taskwarrior makes available. It's possible the
                # modified task could be a recurring task, and to make the
                # individual edit actions consistent, recurrence.confirmation
                # is set to 'no', so that only the edited recurring task is
                # modified.
                returncode, stdout, stderr = self.command.run(['task', 'rc.recurrence.confirmation=no', metadata['uuid'], 'modify', 'wait:%s' % wait], capture_output=True)
                if returncode == 0:
                    self.table.flash_focus()
                    self.update_report()
                    self.activate_message_bar('Task %s wait updated' % self.model.task_id(metadata['uuid']))
                else:
                    self.activate_message_bar("Error setting wait: %s" % stderr, 'error')
            elif op == 'context':
                # TODO: Validation if more than one arg passed.
                context = args[0] if len(args) > 0 else 'none'
                if context != 'none':
                    # In case a new context was added between bootstraps.
                    self.load_contexts()
                if self.execute_command(['task', 'context', context], wait=self.wait):
                    self.activate_message_bar('Context switched to: %s' % context)
                else:
                    self.activate_message_bar('Error switching context', 'error')
            elif len(args) > 0:
                if op == 'add':
                    if self.execute_command(['task', 'add'] + args, wait=self.wait):
                        task = self.task_get_latest()
                        self.activate_message_bar('Task %s added' % task_id_or_uuid_short(task))
                        self.focus_new_task(task)
                elif op == 'modify':
                    # TODO: Will this break if user clicks another list item
                    # before hitting enter?
                    if self.execute_command(['task', metadata['uuid'], 'modify'] + args, wait=self.wait):
                        self.activate_message_bar('Task %s modified' % self.model.task_id(metadata['uuid']))
                elif op == 'annotate':
                    task = self.model.task_annotate(metadata['uuid'], data['text'])
                    if task:
                        self.table.flash_focus()
                        self.update_report()
                        self.activate_message_bar('Annotated task %s' % self.model.task_id(task['uuid']))
                elif op == 'tag':
                    task = self.model.task_tags(metadata['uuid'], args)
                    if task:
                        self.table.flash_focus()
                        self.update_report()
                        self.activate_message_bar('Task %s tags updated' % self.model.task_id(task['uuid']))
                elif op in ('search-forward', 'search-reverse'):
                    self.search_set_term(data['text'])
                    self.search_set_direction(op)
                    self.search(reverse=(op == 'search-reverse'))
        self.widget.focus_position = 'body'
        if 'uuid' in metadata:
            self.task_list.focus_by_task_uuid(metadata['uuid'], self.previous_focus_position)

    def focus_new_task(self, task):
        if self.config.get('vit', 'focus_on_add'):
            self.task_list.focus_by_task_uuid(task['uuid'], self.previous_focus_position)

    def key_pressed(self, key):
        if is_mouse_event(key):
            return None
        keys = self.key_cache.get(key)
        if self.action_manager_registrar.handled_action(keys):
            self.activate_message_bar()
            self.action_manager_registrar.execute_handler(keys)
        elif self.check_macro(keys):
            self.clear_key_cache()
            self.activate_message_bar()
            self.execute_macro(keys)
        elif keys in self.key_cache.multi_key_cache:
            self.key_cache.set(keys)
            self.update_status_key_cache()
        else:
            self.clear_key_cache()

    def on_select(self, row, size, key):
        keys = self.key_cache.get(key)
        if self.action_manager_registrar.handled_action(keys):
            self.activate_message_bar()
            self.action_manager_registrar.execute_handler(keys)
            return None
        else:
            return key

    def activate_help(self, args):
        [self.original_header_top_contents, self.original_header_bottom_contents] = self.header.contents
        self.original_widget_body = self.widget.body
        self.original_footer = self.footer
        help_widget = self.help.update(args)
        self.header.contents[:] = []
        self.widget.body = help_widget
        self.footer = urwid.Pile([])

    def deactivate_help(self, data):
        self.header.contents[:] = [self.original_header_top_contents, self.original_header_bottom_contents]
        self.widget.body = self.original_widget_body
        self.footer = self.original_footer

    def ex(self, text, metadata):
        args = string_to_args(text)
        if len(args):
            command = args.pop(0)
            if command in ('h', 'help'):
                metadata = {}
                self.activate_help(args)
            elif command in ('q',):
                self.quit()
            elif command in ('!', '!r', '!w', '!rw', '!wr'):
                kwargs = {}
                if command in ('!', '!w'):
                    kwargs['update_report'] = False
                if command in ('!', '!r'):
                    kwargs['confirm'] = None
                    kwargs['wait'] = False
                else:
                    kwargs['wait'] = True
                uuid, _ = self.get_focused_task()
                if not uuid:
                    uuid = ""
                kwargs['custom_env'] = {
                    "VIT_TASK_UUID": uuid,
                }
                self.execute_command(args, **kwargs)
            elif command.isdigit():
                self.task_list.focus_by_task_id(int(command))
                if 'uuid' in metadata:
                    metadata.pop('uuid')
            elif command in self.reports:
                self.extra_filters = args
                self.update_report(command)
                if 'uuid' in metadata:
                    metadata.pop('uuid')
            elif command in self.task_config.disallowed_reports:
                self.activate_message_bar("Report '%s' is non-standard, use ':!w task %s'" % (command, command), 'error')
            else:
                # Matches 's/foo/bar/' and s%/foo/bar/, allowing for separators
                # to be any non-word character.
                matches = re.match(r'^%?s(\W)((?:(?!\1).)*)\1((?:(?!\1).)*)\1$', text)
                if matches and 'uuid' in metadata:
                    before, after = matches.group(2, 3)
                    task = self.model.get_task(metadata['uuid'])
                    if task:
                        description = re.sub(r'%s' % before, after, task['description'])
                        task = self.model.task_description(metadata['uuid'], description)
                        if task:
                            self.table.flash_focus()
                            self.update_report()
                            self.activate_message_bar('Task %s description updated' % self.model.task_id(task['uuid']))
        return metadata

    def search_set_term(self, text):
        self.search_term_active = text

    def search_set_direction(self, op):
        self.search_direction_reverse = op == 'search-reverse'

    def search(self, reverse=False):
        if not self.search_term_active:
            return
        self.table.batcher.add(0)
        self.search_display_message(reverse)
        current_index = 0 if self.task_list.focus is None else self.task_list.focus_position
        new_focus = self.search_rows(self.search_term_active, current_index, reverse)
        if new_focus is None:
            self.activate_message_bar("Pattern not found: %s" % self.search_term_active, 'error')
        else:
            self.task_list.focus_position = new_focus

    def search_rows(self, term, start_index=0, reverse=False):
        escaped_term = re.escape(term)
        search_regex = re.compile(escaped_term, re.MULTILINE)
        rows = self.table.rows
        current_index = start_index
        last_index = len(rows) - 1
        if len(rows) > 0:
            start_matches = self.search_row_has_search_term(rows[start_index], search_regex)
            current_index = self.search_increment_index(current_index, reverse)
            while True:
                if reverse and current_index < 0:
                    self.search_loop_warning('TOP', reverse)
                    current_index = last_index
                elif not reverse and current_index > last_index:
                    self.search_loop_warning('BOTTOM', reverse)
                    current_index = 0
                if self.search_row_has_search_term(rows[current_index], search_regex):
                    return current_index
                if current_index == start_index:
                    return start_index if start_matches else None
                current_index = self.search_increment_index(current_index, reverse)

    def search_increment_index(self, current_index, reverse=False):
        return current_index + (-1 if reverse else 1)

    def search_display_message(self, reverse=False):
        self.activate_message_bar("Search %s for: %s" % ('reverse' if reverse else 'forward', self.search_term_active))

    def search_loop_warning(self, hit, reverse=False):
        self.activate_message_bar('Search hit %s, continuing at %s' % (hit, hit == 'TOP' and 'BOTTOM' or 'TOP'))
        self.loop.draw_screen()
        time.sleep(0.8)
        self.search_display_message(reverse)

    def reconstitute_markup_element_as_string(self, accum, markup):
        if isinstance(markup, tuple):
            _, markup = markup
        return accum + markup

    def reconstitute_markup_as_string(self, markup):
        if isinstance(markup, list):
            return reduce(self.reconstitute_markup_element_as_string, markup, '')
        return self.reconstitute_markup_element_as_string('', markup)

    def search_row_has_search_term(self, row, search_regex):
        # TODO: Cleaner way to detect valid searchable row.
        if hasattr(row, 'data'):
            for column in row.data:
                value = self.reconstitute_markup_as_string(column)
                if value and search_regex.search(value):
                    return True
            return False

    def get_focused_task(self):
        try:
            uuid = self.task_list.focus.uuid
            task = self.model.get_task(uuid)
            return uuid, task
        except:
            pass
        return False, False

    def signal_quit(self, signal):
        #import debug
        #debug.file("VIT received %s signal, quitting" % signal)
        self.quit()

    def quit(self):
        self.teardown_pid()
        raise urwid.ExitMainLoop()

    def build_task_table(self):
        self.table = TaskTable(self.config, self.task_config, self.formatter, self.loop.screen, on_select=self.on_select, event=self.event, action_manager=self.action_manager, request_reply=self.request_reply, markers=self.markers, draw_screen_callback=self.loop.draw_screen)

    def update_task_table(self):
        self.table.update_data(self.reports[self.report], self.model.tasks)

    def init_task_list(self):
        self.model = TaskListModel(self.task_config, self.reports)

    def init_autocomplete(self):
        context_list = list(self.contexts.keys()) + ['none']
        self.autocomplete = AutoComplete(self.config, extra_filters={'report': self.reports.keys(), 'help': self.help.autocomplete_entries(), 'context': context_list})

    def init_command_bar(self):
        abort_backspace = self.config.get('vit', 'abort_backspace')
        self.command_bar = CommandBar(autocomplete=self.autocomplete, abort_backspace=abort_backspace, event=self.event)

    def build_frame(self):
        self.status_report = urwid.AttrMap(urwid.Text('Welcome to VIT'), 'status')
        self.status_context = urwid.AttrMap(urwid.Text(''), 'status')
        self.status_performance = urwid.AttrMap(urwid.Text('', align='center'), 'status')
        self.status_version = urwid.AttrMap(urwid.Text('vit (%s)' % version.VIT, align='center'), 'status')
        self.status_tasks_shown = urwid.AttrMap(urwid.Text('', align='right'), 'status')
        self.status_tasks_completed = urwid.AttrMap(urwid.Text('', align='right'), 'status')
        self.top_column_left = urwid.Pile([
            self.status_report,
            self.status_context,
        ])
        self.top_column_center = urwid.Pile([
            self.status_version,
            self.status_performance,
        ])
        self.top_column_right = urwid.Pile([
            self.status_tasks_shown,
            self.status_tasks_completed,
        ])
        self.header = urwid.Pile([
            urwid.Columns([
                self.top_column_left,
                self.top_column_center,
                self.top_column_right,
            ]),
            urwid.Text('Loading...'),
        ])
        self.footer = MultiWidget()
        self.init_autocomplete()
        self.init_command_bar()
        self.message_bar = urwid.Text('', align='center')
        self.footer.add_widget('command', self.command_bar)
        self.footer.add_widget('message', self.message_bar)

    def command_error(self, returncode, error_message):
        if returncode != 0:
            self.activate_message_bar("Command error: %s" % error_message, 'error')
            return True
        return False

    def execute_command(self, args, **kwargs):
        update_report = True
        wait = True
        if 'update_report' in kwargs:
            update_report = kwargs.pop('update_report')
        if 'wait' in kwargs:
            wait = kwargs.pop('wait')
            if not wait:
                kwargs['confirm'] = None
        self.loop.stop()
        # TODO: This is a shitty hack, if not waiting, then we must
        # override the confirmation setting for recurring tasks.
        if not wait and args[0] == 'task':
            args.insert(1, 'rc.recurrence.confirmation=no')
        def execute():
            returncode, output = self.command.result(args, **kwargs)
            if self.command_error(returncode, output):
                return False
            return True
        success = execute()
        if update_report and success:
            self.update_report()
        self.loop.start()
        return success

    def activate_command_bar(self, op, caption, metadata={}, edit_text=None):
        metadata['op'] = op
        self.footer.show_widget('command')
        self.setup_autocomplete(op)
        self.command_bar.activate(caption, metadata, edit_text)
        self.widget.focus_position = 'footer'


    def activate_command_bar_filter(self):
        self.activate_command_bar('filter', 'Filter: ')

    def task_undo(self):
        self.execute_command(['task', 'undo'])

    def task_sync(self):
        self.execute_command(['task', 'sync'])

    def task_get_latest(self):
        returncode, stdout, stderr = self.command.run(['task', '+LATEST', 'uuids'], capture_output=True)
        if returncode == 0:
            return self.model.get_task(stdout)
        else:
            raise RuntimeError("Error retrieving latest task UUID: %s" % stderr)

    def activate_command_bar_quit_with_confirm(self):
        if self.confirm:
            self.activate_command_bar('quit', 'Quit?', {'choices': {'y': True}})
        else:
            self.quit()

    def activate_command_bar_ex(self):
        metadata = {}
        uuid, _ = self.get_focused_task()
        if uuid:
            metadata['uuid'] = uuid
        self.activate_command_bar('ex', ':', metadata)

    def activate_command_bar_ex_read_wait_task(self):
        self.activate_command_bar('ex', ':', {}, edit_text='!rw task ')

    def activate_command_bar_search_forward(self):
        self.activate_command_bar('search-forward', '/', {'history': 'search'})

    def activate_command_bar_search_reverse(self):
        self.activate_command_bar('search-reverse', '?', {'history': 'search'})

    def activate_command_bar_search_next(self):
        self.search(reverse=self.search_direction_reverse)

    def activate_command_bar_search_previous(self):
        self.search(reverse=not self.search_direction_reverse)

    def activate_command_bar_task_context(self):
        self.activate_command_bar('context', 'Context: ')

    def global_escape(self):
        self.denotation_pop_up.close_pop_up()

    def activate_command_bar_add(self):
        self.activate_command_bar('add', 'Add: ')

    def task_done(self, uuid):
        success, task = self.model.task_done(uuid)
        if task:
            if success:
                self.table.flash_focus()
                self.update_report()
                self.activate_message_bar('Task %s marked done' % self.model.task_id(task['uuid']))
            else:
                self.activate_message_bar('Error: %s' % task, 'error')

    def task_delete(self, uuid):
        success, task = self.model.task_delete(uuid)
        if task:
            if success:
                self.table.flash_focus()
                self.update_report()
                self.activate_message_bar('Task %s deleted' % self.model.task_id(task['uuid']))
            else:
                self.activate_message_bar('Error: %s' % task, 'error')

    def task_start_stop(self, uuid):
        success, task = self.model.task_start_stop(uuid)
        if task:
            if success:
                self.table.flash_focus()
                self.update_report()
                self.activate_message_bar('Task %s %s' % (self.model.task_id(task['uuid']), 'started' if task.active else 'stopped'))
            else:
                self.activate_message_bar('Error: %s' % task, 'error')

    def task_action_annotate(self):
        uuid, _ = self.get_focused_task()
        if uuid:
            self.activate_command_bar('annotate', 'Annotate: ', {'uuid': uuid})
            self.task_list.focus_by_task_uuid(uuid, self.previous_focus_position)

    def task_action_delete(self):
        uuid, task = self.get_focused_task()
        if task:
            if self.confirm:
                self.activate_command_bar('delete', 'Delete task %s? (y/n): ' % self.model.task_id(uuid), {'uuid': uuid, 'choices': {'y': True}})
            else:
                self.task_delete(uuid)
                self.task_list.focus_by_task_uuid(uuid, self.previous_focus_position)

    def task_action_denotate(self):
        uuid, task = self.get_focused_task()
        if task and task['annotations']:
            self.denotation_pop_up.open(task)

    def task_action_modify(self):
        uuid, _ = self.get_focused_task()
        if uuid:
            self.activate_command_bar('modify', 'Modify: ', {'uuid': uuid})
            self.task_list.focus_by_task_uuid(uuid, self.previous_focus_position)

    def task_action_start_stop(self):
        uuid, task = self.get_focused_task()
        if task:
            if self.confirm:
                self.activate_command_bar('start-stop', '%s task %s? (y/n): ' % (task.active and 'Stop' or 'Start', self.model.task_id(uuid)), {'uuid': uuid, 'choices': {'y': True}})
            else:
                self.task_start_stop(uuid)
                self.task_list.focus_by_task_uuid(uuid, self.previous_focus_position)

    def task_action_done(self):
        uuid, task = self.get_focused_task()
        if task:
            if self.confirm:
                self.activate_command_bar('done', 'Mark task %s done? (y/n): ' % self.model.task_id(uuid), {'uuid': uuid, 'choices': {'y': True}})
            else:
                self.task_done(uuid)
                self.task_list.focus_by_task_uuid(uuid, self.previous_focus_position)

    def task_action_priority(self):
        uuid, _ = self.get_focused_task()
        if uuid:
            choices = {}
            for choice in self.task_config.priority_values:
                key = choice.lower() or 'n'
                choices[key] = choice
            self.activate_command_bar('priority', 'Priority (%s): ' % '/'.join(choices), {'uuid': uuid, 'choices': choices})

    def task_action_project(self):
        uuid, _ = self.get_focused_task()
        if uuid:
            self.activate_command_bar('project', 'Project: ', {'uuid': uuid})

    def task_action_tags(self):
        uuid, _ = self.get_focused_task()
        if uuid:
            self.activate_command_bar('tag', 'Tag: ', {'uuid': uuid})

    def task_action_wait(self):
        uuid, _ = self.get_focused_task()
        if uuid:
            self.activate_command_bar('wait', 'Wait: ', {'uuid': uuid})

    def task_action_edit(self):
        uuid, _ = self.get_focused_task()
        if uuid:
            self.execute_command(['task', uuid, 'edit'], wait=self.wait)
            self.task_list.focus_by_task_uuid(uuid, self.previous_focus_position)

    def task_action_show(self):
        uuid, _ = self.get_focused_task()
        if uuid:
            self.execute_command(['task', uuid, 'info'], update_report=False)
            self.task_list.focus_by_task_uuid(uuid)

    def get_available_task_columns(self):
        returncode, stdout, stderr = self.command.run(['task', '_columns'], capture_output=True)
        if returncode == 0:
            self.available_columns = stdout.split()
        else:
            raise RuntimeError("Error retrieving available task columns: %s" % stderr)

    def refresh_blocking_task_uuids(self):
        returncode, stdout, stderr = self.command.run(['task', 'uuids', '+BLOCKING'], capture_output=True)
        if returncode == 0:
            self.blocking_task_uuids = stdout.split()
        else:
            raise RuntimeError("Error retrieving blocking task UUIDs: %s" % stderr)

    def setup_autocomplete(self, op):
        callback = self.command_bar.set_edit_text_callback()
        if op in ('filter', 'add', 'modify'):
            self.autocomplete.setup(callback)
        elif op in ('ex',):
            filters = ('report', 'column', 'project', 'tag', 'help')
            filter_config = copy.deepcopy(self.autocomplete.default_filter_config)
            filter_config['report'] = {
                'include_unprefixed': True,
                'root_only': True,
            }
            filter_config['help'] = {
                'include_unprefixed': True,
                'root_only': True,
            }
            self.autocomplete.setup(callback, filters=filters, filter_config=filter_config)
        elif op in ('project',):
            filters = ('project',)
            filter_config = {
                'project': {
                    'prefixes': [],
                    'include_unprefixed': True,
                },
            }
            self.autocomplete.setup(callback, filters=filters, filter_config=filter_config)
        elif op in ('tag',):
            filters = ('tag',)
            filter_config = {
                'tag': {
                    'prefixes': ['+', '-'],
                    'include_unprefixed': True,
                },
            }
            self.autocomplete.setup(callback, filters=filters, filter_config=filter_config)
        elif op in ('context',):
            filters = ('context',)
            filter_config = {
                'context': {
                    'include_unprefixed': True,
                    'root_only': True,
                },
            }
            self.autocomplete.setup(callback, filters=filters, filter_config=filter_config)

    def activate_message_bar(self, message='', message_type='status'):
        self.footer.show_widget('message')
        display = 'message %s' % message_type
        self.message_bar.set_text((display, message))

    def update_status_report(self):
        filtered_report = 'task %s %s' % (self.report, ' '.join(self.extra_filters))
        self.status_report.original_widget.set_text(filtered_report)

    def update_status_performance(self, seconds):
        text = 'Exec. time: %dms' % (seconds * 1000)
        self.status_performance.original_widget.set_text(text)

    # TODO: This is riding on top of status_performance currently, should
    # probably be abstracted
    def update_status_key_cache(self):
        keys = self.key_cache.get()
        text = 'Key cache: %s' % keys if keys else ''
        self.status_performance.original_widget.set_text(text)

    def update_status_context(self):
        text = 'Context: %s' % self.context if self.context else 'No context'
        self.status_context.original_widget.set_text(text)

    def update_status_tasks_shown(self):
        num_tasks = len(self.model.tasks)
        text = '%s %s shown' % (num_tasks, num_tasks == 1 and 'task' or 'tasks')
        self.status_tasks_shown.original_widget.set_text(text)

    def update_status_tasks_completed(self):
        returncode, stdout, stderr = self.command.run(['task', '+COMPLETED', 'count'], capture_output=True)
        if returncode == 0:
            num_tasks = int(stdout.strip())
            text = '%s %s completed' % (num_tasks, num_tasks == 1 and 'task' or 'tasks')
            self.status_tasks_completed.original_widget.set_text(text)
        else:
            raise RuntimeError("Error retrieving completed tasks: %s" % stderr)

    def async_refresh(self, _):
        self.refresh()

    def refresh(self, load_early_config=True):
        self.bootstrap(load_early_config)
        self.build_main_widget()
        # NOTE: Don't see any other way to clear the old palette.
        self.loop.screen._palette = {}
        self.loop.screen.register_palette(self.theme)
        self.loop.screen.clear()
        self.loop.widget = self.widget

    def update_report(self, report=None):
        start = time.time()
        self.task_list = self.table.listbox
        self.previous_focus_position = self.task_list.focus_position if self.task_list.list_walker else 0
        if report:
            self.report = report
        self.set_active_context()
        self.task_config.get_projects()
        self.refresh_blocking_task_uuids()
        self.formatter.recalculate_due_datetimes()
        context_filters = self.contexts[self.context]['filter'] if self.context and self.reports[self.report].get('context', 1) else []
        try:
            self.model.update_report(self.report, context_filters=context_filters, extra_filters=self.extra_filters)
        except VitException as err:
            self.activate_message_bar(str(err), 'error')
            return
        self.update_task_table()
        self.update_status_report()
        self.update_status_context()
        self.update_status_tasks_shown()
        self.update_status_tasks_completed()
        self.header.contents[1] = (self.table.header, self.header.options())
        self.denotation_pop_up = DenotationPopupLauncher(self.task_list, self.formatter, self.loop.screen, event=self.event, request_reply=self.request_reply, action_manager=self.action_manager)
        self.widget.body = self.denotation_pop_up
        self.autocomplete.refresh()
        end = time.time()
        self.update_status_performance(end - start)

    def build_main_widget(self, report=None):
        if report:
            self.report = report
        self.init_task_list()
        self.build_frame()
        self.widget = MainFrame(
            urwid.ListBox([]),
            header=self.header,
            footer=self.footer,
            key_cache=self.key_cache,
            action_manager=self.action_manager,
            refresh=self.refresh,
        )
        self.update_report(self.report)



theme = [
    ('list-header', '', '', '', '', ''),
    ('list-header-column', 'underline', '', '', 'underline', ''),
    ('list-header-column-separator', '', '', '', '', ''),
    ('striped-table-row', 'white', 'dark gray', '', 'white', 'g27'),
    ('reveal focus', 'white', 'dark blue', 'standout', 'white', 'dark blue'),
    ('message status', '', '', '', '', ''),
    ('message error', 'white', 'dark red', 'standout', 'white', 'dark red'),
    ('status', 'dark blue', 'black', '', 'dark blue', 'black'),
    ('flash off', 'black', 'black', 'standout', 'black', 'black'),
    ('flash on', 'white', 'black', 'standout', 'white', 'black'),
    ('pop_up', 'white', 'black', '', 'white', 'black'),
    ('button action', 'white', 'light red', '', 'white', 'light red'),
    ('button cancel', 'black', 'light gray', '', 'black', 'light gray'),
]



import uuid

class ActionManagerRegistrar:
    def __init__(self, registry):
        self.registry = registry
        self.uuid = uuid.uuid4()

    def register(self, name, callback):
        self.registry.register(self.uuid, name, callback)

    def deregister(self, name=None):
        if name:
            self.registry.deregister(self.uuid, name)
        else:
            any(self.registry.deregister(self.uuid, action) for _, action in self.actions().items())

    def actions(self):
        return self.registry.get_registered(self.uuid)

    def execute_handler(self, *args):
        keys, args = args[0], args[1:]
        handled_action = self.handled_action(keys)
        if handled_action:
            handled_action['callback'](*args)
            data = {
                'name': handled_action['name'],
                'args': args,
            }
            self.registry.event.emit('action-manager:action-executed', data)
            return True
        return False

    def handled_action(self, keys):
        keybindings = self.registry.keybindings
        actions = self.actions()
        if keys in keybindings and 'action_name' in keybindings[keys] and keybindings[keys]['action_name'] in actions:
            return actions[keybindings[keys]['action_name']]
        return None

class ActionManagerRegistry:
    def __init__(self, action_registry, keybindings, event=None):
        self.actions = {}
        self.action_registry = action_registry
        self.keybindings = keybindings
        self.event = event

    def get_registrar(self):
        return ActionManagerRegistrar(self)

    def active_registration_id(self, registration_id):
        return registration_id in self.actions

    def get_registered(self, registration_id):
        return self.actions[registration_id] if self.active_registration_id(registration_id) else {}

    def register(self, registration_id, name, callback):
        if not self.active_registration_id(registration_id):
            self.actions[registration_id] = {}
        self.actions[registration_id][self.action_registry.make_action_name(name)] = {
            'name': name,
            'callback': callback,
        }

    def deregister(self, registration_id, name_or_action):
        if self.active_registration_id(registration_id):
            name = name_or_action['name'] if isinstance(name_or_action, dict) else name_or_action
            self.actions[registration_id].pop(self.action_registry.make_action_name(name))




import urwid

from vit import util
from vit.base_list_box import BaseListBox

OVERLAY_WIDTH = 50
OVERLAY_HEIGHT = 15

class AnnotationFrame(urwid.Frame):
    def __init__(self, body, **kwargs):
        self.listbox = body.original_widget
        return super().__init__(body, **kwargs)
    def keypress(self, size, key):
        """Overrides Frame.keypress method.
        """
        if key in ('tab', 'shift tab'):
            self.focus_position = self.focus_position == 'body' and 'footer' or 'body'
            self.listbox.update_focus_blur(self.focus_position == 'body' and 'focus' or 'blur')
            return None
        return super().keypress(size, key)

class AnnotationListBox(BaseListBox):
    """Maps denotation list shortcuts to default ListBox class.
    """

    def __init__(self, body, event=None, request_reply=None, action_manager=None):
        super().__init__(body, event=event, request_reply=request_reply, action_manager=action_manager)
        self.init_event_listeners()

    def list_action_executed(self, size, key):
        data = {
            'size': size,
        }
        self.event.emit('denotation-list:keypress', data)

    def init_event_listeners(self):
        def signal_handler():
            self.update_focus()
        self.modified_signal = urwid.connect_signal(self.list_walker, 'modified', signal_handler)

    def get_selected_annotation(self):
        return self.list_walker[self.focus_position].annotation

    def update_focus(self):
        if self.previous_focus_position != self.focus_position:
            if self.previous_focus_position is not None:
                self.update_focus_attr({}, position=self.previous_focus_position)
        self.update_focus_attr('reveal focus')
        self.previous_focus_position = self.focus_position

    def update_focus_attr(self, attr, position=None):
        attr = attr if isinstance(attr, dict) else {None: attr}
        if position is None:
            position = self.focus_position
        self.list_walker[position].row.set_attr_map(attr)

    def update_focus_blur(self, op):
        attr = 'reveal focus' if op == 'focus' else 'button cancel'
        self.update_focus_attr(attr)

class SelectableRow(urwid.WidgetWrap):
    """Wraps 'urwid.Columns' to make it selectable.
    """

    def __init__(self, annotation, position, widths, formatter, align="left", on_select=None, space_between=2):
        self.annotation = annotation
        self.position = position

        self._columns = urwid.Columns([
            (widths['entry'], urwid.Text(annotation['entry'].strftime(formatter.annotation), align=align)),
            (widths['description'], urwid.Text(annotation['description'], align=align)),
        ], dividechars=space_between)

        self.row = urwid.AttrMap(self._columns, '')

        # Wrap 'urwid.Columns'.
        super().__init__(self.row)

        # A hook which defines the behavior that is executed when a specified key is pressed.
        self.on_select = on_select

    def __repr__(self):
        return "{}(entry={}, description={})".format(self.__class__.__name__,
                                          self.annotation['entry'], self.annotation['description'])

    def selectable(self):
        return True

    def keypress(self, size, key):
        if self.on_select:
            key = self.on_select(self, size, key)
        return key


class DenotationPopUpDialog(urwid.WidgetWrap):
    """A dialog for denotating tasks."""
    # TODO: Is this necessary? Copied from examples.
    signals = ['close']
    def __init__(self, task, listbox, formatter, event=None):
        self.task = task
        self.listbox = listbox
        self.formatter = formatter
        self.event = event
        self.uuid = self.task['uuid']
        denotate_button = urwid.AttrWrap(urwid.Button("Denotate"), '', 'button action')
        cancel_button = urwid.AttrWrap(urwid.Button("Cancel"), '', 'button cancel')
        # TODO: Calculate these dynamically?
        widths = {
            'entry': 10,
            'description': 32,
        }
        annotations = [SelectableRow(a, idx, widths, self.formatter) for idx, a in enumerate(self.task['annotations'])]
        self.listbox.list_walker[:] = annotations
        self.listbox.focus_position = 0
        def denotate(button):
            self._emit("close")
            data = {
                'uuid': self.uuid,
                'annotation': self.listbox.get_selected_annotation(),
            }
            self.event.emit("task:denotate", data)
        urwid.connect_signal(denotate_button.original_widget, 'click', denotate)
        urwid.connect_signal(cancel_button.original_widget, 'click',
            lambda button:self._emit("close"))
        frame = AnnotationFrame(
            urwid.Padding(self.listbox, left=1, right=1),
            header=urwid.Text("Select the annotation, then select 'Denotate'\n"),
            footer=urwid.Columns([urwid.Padding(denotate_button, left=5, right=6), urwid.Padding(cancel_button, left=6, right=5)]),
        )
        padded_frame = urwid.Padding(frame, left=1, right=1)
        box = urwid.LineBox(padded_frame, "Denotate task %s" % util.task_id_or_uuid_short(task))
        super().__init__(urwid.AttrWrap(box, 'pop_up'))

class DenotationPopupLauncher(urwid.PopUpLauncher):
    def __init__(self, original_widget, formatter, screen, event=None, request_reply=None, action_manager=None):
        self.formatter = formatter
        self.screen = screen
        self.event = event
        self.request_reply = request_reply
        self.action_manager = action_manager
        self.listbox = AnnotationListBox(urwid.SimpleFocusListWalker([]), event=self.event, request_reply=self.request_reply, action_manager=self.action_manager)
        super().__init__(original_widget)

    def set_task(self, task):
        self.task = task

    def open(self, task):
        self.set_task(task)
        self.open_pop_up()

    def create_pop_up(self):
        pop_up = DenotationPopUpDialog(self.task, self.listbox, self.formatter, event=self.event)
        urwid.connect_signal(pop_up, 'close',
            lambda button: self.close_pop_up())
        return pop_up

    def get_pop_up_parameters(self):
        screen_width, _ = self.screen.get_cols_rows()
        left = round((screen_width - OVERLAY_WIDTH) / 2)
        return {'left':left, 'top':1, 'overlay_width':OVERLAY_WIDTH, 'overlay_height':OVERLAY_HEIGHT}



from vit.formatter import DateTime

class Until(DateTime):
    def get_until_state(self, until, task):
        return self.formatter.get_until_state(until, task)

    def colorize(self, until, task):
        return self.colorizer.until(self.get_until_state(until, task))



from vit.formatter.description import Description
from vit.util import unicode_len

class DescriptionDesc(Description):
    def format(self, description, task):
        if not description:
            return self.empty()
        colorized_description = self.colorize_description(description)
        return (unicode_len(description), colorized_description)



from shutil import copyfile

import os
import re
import shlex
import datetime

from functools import reduce

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

from vit import env, xdg
from vit.process import Command

SORT_ORDER_CHARACTERS = ['+', '-']
SORT_COLLATE_CHARACTERS = ['/']
VIT_CONFIG_FILE = 'config.ini'
FILTER_EXCLUSION_REGEX = re.compile(r'^limit:')
FILTER_PARENS_REGEX = re.compile(r'([\(\)])')
CONFIG_BOOLEAN_TRUE_REGEX = re.compile(r'1|yes|true', re.IGNORECASE)
# TaskParser expects clean hierarchies in the Taskwarrior dotted config names.
# However, this is occasionally violated, with a leaf ending in both a string
# value and another branch. The below list contains the config values that
# violate this convention, and transform them into a single additional branch
# of value CONFIG_STRING_LEAVES_DEFAULT_BRANCH
CONFIG_STRING_LEAVES = [
    'color.calendar.due',
    'color.due',
    'color.label',
    'dateformat',
]
CONFIG_STRING_LEAVES_DEFAULT_BRANCH = 'default'

DEFAULTS = {
    'taskwarrior': {
        'taskrc': '~/.taskrc',
    },
    'vit': {
        'default_keybindings': 'vi',
        'theme': 'default',
        'confirmation': True,
        'wait': True,
        'mouse': False,
        'abort_backspace': False,
        'focus_on_add': False,
        'pid_dir': '',
        'flash_focus_repeat_times': 2,
        'flash_focus_pause_seconds': 0.1,
    },
    'report': {
        'default_report': 'next',
        'default_filter_only_report': 'next',
        'indent_subprojects': True,
        'row_striping': True,
    },
    'marker': {
        'enabled': True,
        'header_label': '',
        'columns': 'all',
        'require_color': True,
        'include_subprojects': True,
    },
    'color': {
        'enabled': True,
        'include_subprojects': True,
        'none_label': '[NONE]',
    },
}

# strftime() works differently on Windows, test here.
test_datetime = datetime.datetime(1900, 1, 1)
NO_PAD_FORMAT_CODE = '-' if test_datetime.strftime('%-d') == '1' else '#' if test_datetime.strftime('%#d') == '1' else ''

DATE_FORMAT_MAPPING = {
    'm': '%%%sm' % NO_PAD_FORMAT_CODE,   # 1 or 2 digit month number, eg '1', '12'
    'M': '%m',                           # 2 digit month number, eg '01', '12'
    'd': '%%%sd' % NO_PAD_FORMAT_CODE,   # 1 or 2 digit day of month number¸ eg '1', '12'
    'D': '%d',                           # 2 digit day of month number, eg '01', '30'
    'y': '%y',                           # 2 digit year, eg '12', where the century is assumed to be '20', therefore '2012'
    'Y': '%Y',                           # 4 digit year, eg '2015'
    'h': '%%%sH' % NO_PAD_FORMAT_CODE,   # 1 or 2 digit hours, eg '1', '23'
    'H': '%H',                           # 2 digit hours, eg '01', '23'
    'n': '%%%sM' % NO_PAD_FORMAT_CODE,   # 1 or 2 digit minutes, eg '1', '59'
    'N': '%M',                           # 2 digit minutes, eg '01', '59'
    's': '%%%sS' % NO_PAD_FORMAT_CODE,   # 1 or 2 digit seconds, eg '1', '59'
    'S': '%S',                           # 2 digit seconds, eg '01', '59'
    'v': '%%%sU' % NO_PAD_FORMAT_CODE,   # 1 or 2 digit week number, eg '1', '52'
    'V': '%U',                           # 2 digit week number, eg '01', '52'
    'a': '%a',                           # 3-character English day name abbreviation, eg 'mon', 'tue'
    'A': '%A',                           # Complete English day name, eg 'monday', 'tuesday'
    'b': '%b',                           # 3-character English month name abbreviation, eg 'jan', 'feb'
    'B': '%B',                           # Complete English month name, eg 'january', 'february'
    'j': '%%%sj' % NO_PAD_FORMAT_CODE,   # 1, 2 or 3 digit day-of-year number, sometimes referred to as a Julian date, eg '1', '11', or '365'
    'J': '%j',                           # 3 digit day of year number, sometimes referred to as a Julian date, eg '001', '011', or '365'
}

class ConfigParser:
    def __init__(self, loader):
        self.loader = loader
        self.config = configparser.SafeConfigParser()
        self.config.optionxform = str
        self.user_config_dir = self.loader.user_config_dir
        self.user_config_filepath = '%s/%s' % (self.user_config_dir, VIT_CONFIG_FILE)
        if not self.config_file_exists(self.user_config_filepath):
            self.optional_create_config_file(self.user_config_filepath)
        self.config.read(self.user_config_filepath)
        self.taskrc_path = self.get_taskrc_path()
        self.validate_taskrc()
        self.defaults = DEFAULTS
        self.set_config_data()

    def set_config_data(self):
        self.subproject_indentable = self.is_subproject_indentable()
        self.row_striping_enabled = self.is_row_striping_enabled()
        self.confirmation_enabled = self.is_confirmation_enabled()
        self.wait_enabled = self.is_wait_enabled()
        self.mouse_enabled = self.is_mouse_enabled()

    def validate_taskrc(self):
        try:
            open(self.taskrc_path, 'r').close()
        except FileNotFoundError:
            message = """
%s not found.

VIT requires a properly configured TaskWarrior instance in the current
environment. Execute the 'task' binary with no arguments to initialize a new
configuration.
""" % (self.taskrc_path)
            print(message)
            exit(1)

    def config_file_exists(self, filepath):
        try:
            open(filepath, 'r').close()
            return True
        except FileNotFoundError:
            return False

    def optional_create_config_file(self, filepath):
        prompt = "%s doesn't exist, create? (y/n): " % filepath
        try:
            answer = input(prompt)
        except:
            answer = raw_input(prompt)
        if answer in ['y', 'Y']:
            self.create_config_file(filepath)
            prompt = "\n%s created. This is the default user configuration file. \n\nIt is heavily commented with explanations and lists the default values for all available user configuration variables. Check it out!\n\nPress enter to continue..." % filepath
            try:
                input(prompt)
            except:
                raw_input(prompt)

    def create_config_file(self, filepath):
        dirname = os.path.dirname(filepath)
        try:
            os.mkdir(dirname)
        except FileExistsError:
            pass
        basedir = os.path.dirname(os.path.realpath(__file__))
        copyfile('%s/config/config.sample.ini' % basedir, filepath)

    def get(self, section, key):
        default = DEFAULTS[section][key]
        try:
            value = self.config.get(section, key)
            return self.transform(key, value, default)
        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
            return default

    def items(self, section):
        try:
            return self.config.items(section)
        except configparser.NoSectionError:
            return []

    def has_section(self, section):
        return self.config.has_section(section)

    def transform(self, key, value, default):
        if isinstance(default, bool):
            return self.transform_bool(value)
        elif isinstance(default, int):
            return self.transform_int(value)
        elif isinstance(default, float):
            return self.transform_float(value)
        else:
            return value

    def transform_bool(self, value):
        return True if CONFIG_BOOLEAN_TRUE_REGEX.match(value) else False

    def transform_int(self, value):
        return int(value)

    def transform_float(self, value):
        return float(value)

    def get_taskrc_path(self):
        taskrc_path = os.path.expanduser('TASKRC' in env.user and env.user['TASKRC'] or self.get('taskwarrior', 'taskrc'))

        if not os.path.exists(taskrc_path):
            xdg_dir = xdg.get_xdg_config_dir(taskrc_path, "task")
            if xdg_dir:
                taskrc_path = os.path.join(xdg_dir, "taskrc")

        return taskrc_path

    def is_subproject_indentable(self):
        return self.get('report', 'indent_subprojects')

    def is_row_striping_enabled(self):
        return self.get('report', 'row_striping')

    def is_confirmation_enabled(self):
        return self.get('vit', 'confirmation')

    def is_wait_enabled(self):
        return self.get('vit', 'wait')

    def is_mouse_enabled(self):
        return self.get('vit', 'mouse')

    def get_flash_focus_repeat_times(self):
        return self.get('vit', 'flash_focus_repeat_times')

    def get_flash_focus_pause_seconds(self):
        return self.get('vit', 'flash_focus_pause_seconds')

class TaskParser:
    def __init__(self, config):
        self.config = config
        self.task_config = []
        self.projects = []
        self.contexts = {}
        self.reports = {}
        self.disallowed_reports = [
            'timesheet',
        ]
        self.command = Command(self.config)
        self.get_task_config()
        self.get_projects()
        self.set_config_data()

    def set_config_data(self):
        self.print_empty_columns = self.is_truthy(self.subtree('print.empty.columns'))
        self.priority_values = self.get_priority_values()

    def get_task_config(self):
        self.task_config = []
        returncode, stdout, stderr = self.command.run('task _show', capture_output=True)
        if returncode == 0:
            lines = list(filter(lambda x: True if x else False, stdout.split("\n")))
            for line in lines:
                hierarchy, values = line.split('=', maxsplit=1)
                self.task_config.append((hierarchy, values))
        else:
            raise RuntimeError('Error parsing task config: %s' % stderr)

    def get_active_context(self):
        returncode, stdout, stderr = self.command.run('task _get rc.context', capture_output=True)
        if returncode == 0:
            return stdout.strip()
        else:
            raise RuntimeError('Error retrieving active context: %s' % stderr)

    def get_projects(self):
        returncode, stdout, stderr = self.command.run('task _projects', capture_output=True)
        if returncode == 0:
            self.projects = stdout.split("\n")
            # Ditch the trailing newline.
            self.projects.pop()
        else:
            raise RuntimeError('Error parsing task projects: %s' % stderr)

    def get_priority_values(self):
        return self.subtree('uda.priority.values').split(',')

    def transform_string_leaves(self, hierarchy):
        if hierarchy in CONFIG_STRING_LEAVES:
            hierarchy += '.%s' % CONFIG_STRING_LEAVES_DEFAULT_BRANCH
        return hierarchy

    def filter_to_dict(self, matcher_regex):
        lines = self.filter(matcher_regex)
        return {key: value for key, value in lines}

    def filter(self, matcher_regex):
        return list(filter(lambda config_pair: re.match(matcher_regex, config_pair[0]), self.task_config))

    def subtree(self, matcher, walk_subtree=True):
        matcher_regex = matcher
        if walk_subtree:
            matcher_regex = r'%s' % (('^%s' % matcher).replace(r'.', r'\.'))
        full_tree = {}
        lines = self.filter(matcher_regex)
        for (hierarchy, value) in lines:
            # NOTE: This is necessary in order to convert Taskwarrior's dotted
            # config syntax into a full tree, as some leaves are both branches
            # and leaves.
            hierarchy = self.transform_string_leaves(hierarchy)
            parts = hierarchy.split('.')
            tree_location = full_tree
            while True:
                if len(parts):
                    part = parts.pop(0)
                    if part not in tree_location:
                        tree_location[part] = {} if len(parts) else value
                    tree_location = tree_location[part]
                else:
                    break
        if walk_subtree:
            parts = matcher.split('.')
            subtree = full_tree
            while True:
                if len(parts):
                    part = parts.pop(0)
                    if part in subtree:
                        subtree = subtree[part]
                else:
                    return subtree
        else:
            return full_tree

    def parse_sort_column(self, column_string):
        order = collate = None
        parts = list(column_string)
        while True:
            if len(parts):
                letter = parts.pop()
                if letter in SORT_ORDER_CHARACTERS:
                    order = letter == '+' and 'ascending' or 'descending'
                elif letter in SORT_COLLATE_CHARACTERS:
                    collate = True
                else:
                    parts.append(letter)
                    break
            else:
                break
        column = ''.join(parts)
        return (column, order, collate)

    def translate_date_markers(self, string):
        return reduce(lambda accum, code: accum.replace(code[0], code[1]), list(DATE_FORMAT_MAPPING.items()), string)

    def get_contexts(self):
        contexts = {}
        self.get_task_config()
        subtree = self.subtree('context.')
        for context, filters in list(subtree.items()):
            contexts[context] = {
                'filter': self.parse_context_filters(context, filters),
            }
        self.contexts = contexts
        return self.contexts

    def parse_context_filters(self, context, filters):
        # Filters can be a string (pre-2.6.0 definition, context.work=+work)
        # or a dict (2.6.0 and newer, context.work.read=+work and
        # context.work.write=+work).
        if type(filters) is dict:
            if 'read' in filters:
                filters = filters['read']
            else:
                return []  # Only contexts with read component defined should be considered.
        filters = shlex.split(re.sub(FILTER_PARENS_REGEX, r' \1 ', filters))
        final_filters = [f for f in filters if not FILTER_EXCLUSION_REGEX.match(f)]
        return final_filters

    def get_reports(self):
        reports = {}
        subtree = self.subtree('report.')
        for report, attrs in list(subtree.items()):
            if report in self.disallowed_reports:
                continue
            reports[report] = {
                'name': report,
                'subproject_indentable': False,
            }
            if 'columns' in attrs:
                reports[report]['columns'] = attrs['columns'].split(',')
            if 'context' in attrs:
                reports[report]['context'] = int(attrs['context'])
            if 'description' in attrs:
                reports[report]['description'] = attrs['description']
            if 'filter' in attrs:
                # Allows quoted strings.
                # Adjust for missing spaces around parentheses.
                filters = shlex.split(re.sub(FILTER_PARENS_REGEX, r' \1 ', attrs['filter']))
                reports[report]['filter'] = [f for f in filters if not FILTER_EXCLUSION_REGEX.match(f)]
            if 'labels' in attrs:
                reports[report]['labels'] = attrs['labels'].split(',')
            else:
                reports[report]['labels'] = [column.title() for column in attrs['columns'].split(',')]
            if 'sort' in attrs:
                columns = attrs['sort'].split(',')
                reports[report]['sort'] = [self.parse_sort_column(c) for c in columns]
            if 'dateformat' in attrs:
                reports[report]['dateformat'] = self.translate_date_markers(attrs['dateformat'])

            self.reports = reports
            # Another pass is needed after all report data has been parsed.
            for report_name, report in self.reports.items():
                self.reports[report_name] = self.rectify_report(report_name, report)
        return self.reports

    def rectify_report(self, report_name, report):
        report['subproject_indentable'] = self.has_project_column(report_name) and self.has_primary_project_ascending_sort(report)
        return report

    def is_truthy(self, value):
        value = str(value)
        return value.lower() in ['y', 'yes', 'on', 'true', '1']

    def has_project_column(self, report_name):
        return self.get_column_index(report_name, 'project') is not None

    def has_primary_project_ascending_sort(self, report):
        try:
            primary_sort = report['sort'][0]
        except KeyError:
            return False

        return primary_sort[0] == 'project' and primary_sort[1] == 'ascending'

    def get_column_index(self, report_name, column):
        return self.reports[report_name]['columns'].index(column) if column in self.reports[report_name]['columns'] else None

    def get_column_label(self, report_name, column):
        column_index = self.get_column_index(report_name, column)
        return self.reports[report_name]['labels'][column_index]



import uuid

class ActionRegistrar:
    def __init__(self, registry):
        self.registry = registry
        self.uuid = uuid.uuid4()

    def register(self, name, description):
        self.registry.register(self.uuid, name, description)

    def deregister(self, name=None):
        if name:
            self.registry.deregister(name)
        else:
            any(self.registry.deregister(action) for _, action in self.actions().items())

    def actions(self):
        return self.registry.get_registered(self.uuid)

class ActionRegistry:
    def __init__(self):
        self.actions = {}
        self.noop_action_name = 'NOOP'

    def get_registrar(self):
        return ActionRegistrar(self)

    def get_registered(self, registration_id):
        return list(filter(lambda action: self.actions[action]['registration_id'] == registration_id, self.actions))

    def register(self, registration_id, name, description):
        self.actions[self.make_action_name(name)] = {
            'name': name,
            'registration_id': registration_id,
            'description': description,
        }

    def deregister(self, name_or_action):
        name = name_or_action['name'] if isinstance(name_or_action, dict) else name_or_action
        self.actions.pop(self.make_action_name(name))

    def get_actions(self):
        return self.actions.keys()

    def make_action_name(self, name):
        return 'ACTION_%s' % name

    def noop(self):
        pass

class RequestReply:
    def __init__(self):
        self.handlers = {}

    def set_handler(self, name, description, callback):
        self.handlers[name] = {
            'description': description,
            'callback': callback,
        }

    def unset_handler(self, name):
        self.handlers.pop(name)

    def request(self, name, *args):
        try:
            return self.handlers[name]['callback'](*args)
        except KeyError:
            raise KeyError("No handler '%s' has been set" % name)



import os
import re
import subprocess
import copy

from vit import env
from vit.util import clear_screen, string_to_args

DEFAULT_CONFIRM = 'Press Enter to continue...'

class Command:

    def __init__(self, config):
        self.config = config
        self.env = env.user.copy()
        self.env['TASKRC'] = self.config.taskrc_path

    def run(self, command, capture_output=False, custom_env={}):
        if isinstance(command, str):
            command = string_to_args(command)
        env = copy.copy(self.env)
        env.update(custom_env)
        kwargs = {
            'env': env,
        }
        if capture_output:
            kwargs.update({
                'stdout': subprocess.PIPE,
                'stderr': subprocess.PIPE,
                'universal_newlines': True,
            })
        try:
            proc = subprocess.Popen(command, **kwargs)
            stdout, stderr = proc.communicate()
            returncode = proc.returncode
        except Exception as e:
            stdout = ''
            stderr = e.message if hasattr(e, 'message') else str(e)
            returncode = 1
        return returncode, stdout, self.filter_errors(returncode, stderr)

    def result(self, command, confirm=DEFAULT_CONFIRM, capture_output=False, print_output=False, clear=True, custom_env={}):
        if clear:
            clear_screen()
        returncode, stdout, stderr = self.run(command, capture_output, custom_env)
        output = returncode == 0 and stdout or stderr
        if print_output:
            print(output)
        if confirm:
            try:
                input(confirm)
            except:
                raw_input(confirm)
        if clear:
            clear_screen()
        return returncode, output

    def filter_errors(self, returncode, error_string):
        if not error_string:
            return '' if returncode == 0 else 'unknown'
        regex = '(TASKRC override)|(^$)'
        filtered_lines = list(filter(lambda s: False if len(re.findall(regex, s)) else True, error_string.split("\n")))
        return "\n".join(filtered_lines)



import re
import urwid
from functools import cmp_to_key, wraps

from vit.color_mappings import task_256_to_urwid_256, task_bright_to_color

VALID_COLOR_MODIFIERS = [
    'bold',
    'underline',
]

INVALID_COLOR_MODIFIERS = [
    'inverse',
]

class TaskColorConfig:
    """Colorized task output.
    """
    def __init__(self, config, task_config, theme, theme_alt_backgrounds):
        self.config = config
        self.task_config = task_config
        self.theme = theme
        self.theme_alt_backgrounds = theme_alt_backgrounds
        self.include_subprojects = self.config.get('color', 'include_subprojects')
        self.task_256_to_urwid_256 = task_256_to_urwid_256()
        # NOTE: Because Taskwarrior disables color on piped commands, and I don't
        # see any portable way to get output from a system command in Python
        # without pipes, the 'color' config setting in Taskwarrior is not used, and
        # instead a custom setting is used.
        self.color_enabled = self.config.get('color', 'enabled')
        self.display_attrs_available, self.display_attrs = self.convert_color_config(self.task_config.filter_to_dict(r'^color\.'))
        self.project_display_attrs = self.get_project_display_attrs()
        if self.include_subprojects:
            self.add_project_children()
        self.inject_alt_background_display_attrs()

    def inject_alt_background_display_attrs(self):
        for display_attr in self.display_attrs.copy():
            name, display_attr_foreground_16, display_attr_background_16, _mono, display_attr_foreground_256, display_attr_background_256 = display_attr
            for modifier, alt_backgrounds in self.theme_alt_backgrounds.items():
                display_attr_modifier = name + modifier
                if self.has_display_attr(name):
                    self.display_attrs_available[display_attr_modifier] = True
                    alt_background_16, alt_background_256 = alt_backgrounds
                    new_background_16 = alt_background_16 if display_attr_background_16 == '' else display_attr_background_16
                    new_background_256 = alt_background_256 if display_attr_background_256 == '' else display_attr_background_256
                    self.display_attrs.append(self.make_display_attr(display_attr_modifier, display_attr_foreground_256, new_background_256, foreground_16=display_attr_foreground_16, background_16=new_background_16))
                else:
                    self.display_attrs_available[display_attr_modifier] = False

    def add_project_children(self):
        color_prefix = 'color.project.'
        for (display_attr, fg16, bg16, m, fg256, bg256) in self.project_display_attrs:
            for entry in self.task_config.projects:
                attr = '%s%s' % (color_prefix, entry)
                if not self.has_display_attr(attr) and attr.startswith('%s.' % display_attr):
                    self.display_attrs_available[attr] = True
                    self.display_attrs.append((attr, fg16, bg16, m, fg256, bg256))

    def has_display_attr(self, display_attr):
        return display_attr in self.display_attrs_available and self.display_attrs_available[display_attr]

    def get_project_display_attrs(self):
        return sorted([(a, fg16, bg16, m, fg256, bg256) for (a, fg16, bg16, m, fg256, bg256) in self.display_attrs if self.display_attrs_available[a] and self.is_project_display_attr(a)], reverse=True)

    def is_project_display_attr(self, display_attr):
        return display_attr[0:14] == 'color.project.'

    def convert_color_config(self, color_config):
        display_attrs_available = {}
        display_attrs = []
        for key, value in color_config.items():
            foreground, background = self.convert_colors(value)
            available = self.has_color_config(foreground, background)
            display_attrs_available[key] = available
            if available:
                display_attrs.append(self.make_display_attr(key, foreground, background))
        return display_attrs_available, display_attrs

    def make_display_attr(self, display_attr, foreground, background, foreground_16=None, background_16=None):
        # TODO: 256 colors need to be translated down to 16 color mode.
        foreground_16 = '' if foreground_16 is None else foreground_16
        background_16 = '' if background_16 is None else background_16
        return (display_attr, foreground_16, background_16, '', foreground, background)

    def has_color_config(self, foreground, background):
        return foreground != '' or background != ''

    def convert_colors(self, color_config):
        # TODO: Maybe a fancy regex eventually...
        color_config = task_bright_to_color(color_config).strip()
        starts_with_on = color_config[0:3] == 'on '
        parts = list(map(lambda p: p.strip(), color_config.split('on ')))
        foreground, background = (parts[0], parts[1]) if len(parts) > 1 else (None, parts[0]) if starts_with_on else (parts[0], None)
        foreground_parts, background_parts = self.make_color_parts(foreground, background)
        return self.convert_color_parts(foreground_parts), self.convert_color_parts(background_parts)

    def convert_color_parts(self, color_parts):
        sorted_parts = self.sort_color_parts(color_parts)
        remapped_colors = self.map_named_colors(sorted_parts)
        return ','.join(remapped_colors)

    def check_invalid_color_parts(self, color_parts):
        invalid_color_parts = {*color_parts} & {*INVALID_COLOR_MODIFIERS}
        if invalid_color_parts:
            raise ValueError("The following TaskWarrior color definitions are unsupported in VIT: %s -- read the documentation for possible workarounds" % ", ".join(invalid_color_parts))

    def map_named_colors(self, color_parts):
        if len(color_parts) > 0 and color_parts[0] in self.task_256_to_urwid_256:
            color_parts[0] = self.task_256_to_urwid_256[color_parts[0]]
        return color_parts

    def make_color_parts(self, foreground, background):
        foreground_parts = self.split_color_parts(foreground)
        self.check_invalid_color_parts(foreground_parts)
        background_parts = self.split_color_parts(background)
        self.check_invalid_color_parts(background_parts)
        return foreground_parts, background_parts

    def split_color_parts(self, color_parts):
        parts = color_parts.split() if color_parts else []
        return parts

    def is_modifier(self, elem):
        return elem in VALID_COLOR_MODIFIERS

    def sort_color_parts(self, color_parts):
        def comparator(first, second):
            if self.is_modifier(first) and not self.is_modifier(second):
                return 1
            elif not self.is_modifier(first) and self.is_modifier(second):
                return -1
            else:
                return 0
        return sorted(color_parts, key=cmp_to_key(comparator))

class TaskColorizer:
    class Decorator:
        def color_enabled(func):
            @wraps(func)
            def verify_color_enabled(self, *args, **kwargs):
                return func(self, *args, **kwargs) if self.color_enabled else None
            return verify_color_enabled
    def __init__(self, color_config):
        self.color_config = color_config
        self.color_enabled = self.color_config.color_enabled
        self.theme_alt_backgrounds = self.color_config.theme_alt_backgrounds
        self.background_modifier = ''
        self.init_keywords()

    def init_keywords(self):
        try:
            self.keywords = self.color_config.task_config.subtree('color.')['keyword']
            self.any_keywords_regex = re.compile('(%s)' % '|'.join(self.keywords.keys()))
        except KeyError:
            self.keywords = []
            self.any_keywords_regex = None

    def has_keywords(self, text):
        return self.any_keywords_regex and self.any_keywords_regex.search(text)

    def extract_keyword_parts(self, text):
        if self.has_keywords(text):
            parts = self.any_keywords_regex.split(text)
            first_part = parts.pop(0)
            return first_part, parts
        return None, None

    def set_background_modifier(self, modifier=''):
        self.background_modifier = modifier if modifier in self.theme_alt_backgrounds else ''

    def add_background_modifier(self, display_attr):
        return display_attr + self.background_modifier

    def make_display_attr(self, display_attr):
        return self.add_background_modifier(display_attr)

    def get_display_attr(self, display_attr):
        return self.make_display_attr(display_attr) if self.color_config.has_display_attr(display_attr) else None

    @Decorator.color_enabled
    def project_none(self):
        return self.get_display_attr('color.project.none')

    @Decorator.color_enabled
    def project(self, project):
        return self.get_display_attr('color.project.%s' % project)

    @Decorator.color_enabled
    def tag_none(self):
        return self.get_display_attr('color.tag.none')

    @Decorator.color_enabled
    def tag(self, tag):
        custom_value = 'color.tag.%s' % tag
        if self.color_config.has_display_attr(custom_value):
            return self.make_display_attr(custom_value)
        elif self.color_config.has_display_attr('color.tagged'):
            return self.make_display_attr('color.tagged')
        return None

    @Decorator.color_enabled
    def uda_none(self, name):
        return self.get_display_attr('color.uda.%s.none' % name)

    @Decorator.color_enabled
    def uda_common(self, name, value):
        custom_value = 'color.uda.%s' % name
        if self.color_config.has_display_attr(custom_value):
            return self.make_display_attr(custom_value)
        elif self.color_config.has_display_attr('color.uda'):
            return self.make_display_attr('color.uda')
        return None

    @Decorator.color_enabled
    def uda_string(self, name, value):
        if not value:
            return self.uda_none(name)
        else:
            custom_value = 'color.uda.%s.%s' % (name, value)
            if self.color_config.has_display_attr(custom_value):
                return self.make_display_attr(custom_value)
            return self.uda_common(name, value)

    @Decorator.color_enabled
    def uda_numeric(self, name, value):
        return self.uda_string(name, value)

    @Decorator.color_enabled
    def uda_duration(self, name, value):
        return self.uda_string(name, value)

    @Decorator.color_enabled
    def uda_date(self, name, value):
        # TODO: Maybe some special string indicators here?
        return self.uda_common(name, value) if value else self.uda_none(name)

    @Decorator.color_enabled
    def uda_indicator(self, name, value):
        return self.uda_string(name, value)

    @Decorator.color_enabled
    def keyword(self, text):
        return self.get_display_attr('color.keyword.%s' % text)

    @Decorator.color_enabled
    def blocking(self):
        return self.get_display_attr('color.blocking')

    @Decorator.color_enabled
    def due(self, state):
        return self.get_display_attr('color.%s' % state) if state else None

    @Decorator.color_enabled
    def status(self, status):
        if status == 'completed' or status == 'deleted':
            value = 'color.%s' % status
            if self.color_config.has_display_attr(value):
                return self.make_display_attr(value)
        return None

    @Decorator.color_enabled
    def blocked(self, depends):
        return self.get_display_attr('color.blocked')

    @Decorator.color_enabled
    def active(self, active):
        return self.get_display_attr('color.active') if active else None

    @Decorator.color_enabled
    def recurring(self, recur):
        return self.get_display_attr('color.recurring')

    @Decorator.color_enabled
    def scheduled(self, scheduled):
        return self.get_display_attr('color.scheduled') if scheduled else None

    @Decorator.color_enabled
    def until(self, until):
        return self.get_display_attr('color.until') if until else None
